@@ -13,6 +13,9 @@
 # limitations under the License.
 
 import json, copy, re
+from kfp_tekton.compiler._k8s_helper import sanitize_k8s_name
+
+from kfp_tekton.tekton import LOOP_GROUP_NAME_LENGTH
 
 
 def _handle_tekton_pipeline_variables(pipeline_run):
@@ -37,6 +40,8 @@ def _handle_tekton_pipeline_variables(pipeline_run):
     for task in task_list:
         if task.get('taskRef', {}):
             continue
+        if 'taskSpec' in task and 'apiVersion' in task['taskSpec']:
+            continue
         for key, val in pipeline_variables.items():
             task_str = json.dumps(task['taskSpec']['steps'])
             if val in task_str:
@@ -92,12 +97,27 @@ def _handle_tekton_custom_task(custom_task: dict, workflow: dict, recursive_task
                 task_dependencies = [dependency['runAfter']]
                 for dep_task in task.get('runAfter', []):
                     # should exclude the custom task itself for cases like graph
-                    dep_task_with_prefix = '-'.join(group_names[:-1] + [dep_task])
+                    dep_task_trim = copy.copy(dep_task)
+                    if len(group_names[-1]) <= LOOP_GROUP_NAME_LENGTH:
+                        dep_task_trim = sanitize_k8s_name(dep_task, max_length=LOOP_GROUP_NAME_LENGTH, rev_truncate=True)
+                    dep_task_with_prefix = '-'.join(group_names[:-1] + [dep_task_trim])
                     if dep_task_with_prefix == dependency['runAfter']:
                         continue
                     if dep_task not in custom_task[dependency['runAfter']]['task_list']:
                         task_dependencies.append(dep_task)
                 task['runAfter'] = task_dependencies
+
+    # process recursive tasks to match parameters
+    for task in recursive_tasks:
+        recursive_graph = custom_task.get(task['taskRef']['name'], {})
+        if recursive_graph:
+            if recursive_graph['spec']['params']:
+                recursive_graph['spec']['params'] = sorted(recursive_graph['spec']['params'], key=lambda k: k['name'])
+            for param in recursive_graph['spec']['params']:
+                recursive_params = [param['name'] for param in task['params']]
+                if param['name'] not in recursive_params:
+                    task['params'].append({'name': param['name'], 'value': "$(params.%s)" % param['name']})
+
     # get custom tasks
     for custom_task_key in custom_task.keys():
         denpendency_list = custom_task[custom_task_key]['spec'].get('runAfter', [])
@@ -111,7 +131,12 @@ def _handle_tekton_custom_task(custom_task: dict, workflow: dict, recursive_task
                 run_after_task_list = []
                 for run_after_task in task.get('runAfter', []):
                     for recursive_task in recursive_tasks:
-                        if recursive_task['name'] in run_after_task and '-'.join(group_names[:-1]) not in run_after_task:
+                        # The subset of the loop group name should be LOOP_GROUP_NAME_LENGTH minus 4 because the
+                        # numbers of loop cannot exceed 1000 due to ETCD limitation.
+                        if sanitize_k8s_name(recursive_task['name'], max_length=(LOOP_GROUP_NAME_LENGTH - 4), rev_truncate=True) \
+                            in run_after_task and '-'.join(group_names[:-1]) not in run_after_task:
+                            if len(group_names[-1]) <= LOOP_GROUP_NAME_LENGTH:
+                                run_after_task = sanitize_k8s_name(run_after_task, max_length=LOOP_GROUP_NAME_LENGTH, rev_truncate=True)
                             run_after_task = '-'.join(group_names[:-1] + [run_after_task])
                             break
                     if run_after_task not in denpendency_list:
@@ -135,7 +160,7 @@ def _handle_tekton_custom_task(custom_task: dict, workflow: dict, recursive_task
                     "params": [{
                         "name": parm['name'],
                         'type': 'string'
-                    } for parm in custom_task[custom_task_key]['spec']['params']],
+                    } for parm in sorted(custom_task[custom_task_key]['spec']['params'], key=lambda k: k['name'])],
                     "tasks": custom_task_cr_tasks
                 }
             }
@@ -188,10 +213,17 @@ def _handle_tekton_custom_task(custom_task: dict, workflow: dict, recursive_task
     # handle the nested custom task case
     # Need to be verified: nested custom task with tasks result as parameters
     nested_custom_tasks = []
-    custom_task_crs_namelist = [custom_task_key for custom_task_key in custom_task.keys()]
+    custom_task_crs_namelist = []
+    for custom_task_key in custom_task.keys():
+        if len(group_names[-1]) <= LOOP_GROUP_NAME_LENGTH:
+            sanitize_k8s_name(custom_task_key, max_length=LOOP_GROUP_NAME_LENGTH, rev_truncate=True)
+        custom_task_crs_namelist.append(custom_task_key)
     for custom_task_key in custom_task.keys():
         for inner_task_name in custom_task[custom_task_key]['task_list']:
-            inner_task_cr_name = '-'.join(group_names[:-1] + [inner_task_name])
+            inner_task_name_trimmed = copy.copy(inner_task_name)
+            if len(group_names[-1]) <= LOOP_GROUP_NAME_LENGTH:
+                inner_task_name_trimmed = sanitize_k8s_name(inner_task_name, max_length=LOOP_GROUP_NAME_LENGTH, rev_truncate=True)
+            inner_task_cr_name = '-'.join(group_names[:-1] + [inner_task_name_trimmed])
             if inner_task_cr_name in custom_task_crs_namelist:
                 nested_custom_tasks.append({
                     "father_ct": custom_task_key,
@@ -221,8 +253,6 @@ def _handle_tekton_custom_task(custom_task: dict, workflow: dict, recursive_task
                 custom_task_cr['spec']['pipelineSpec']['params'].extend([
                     {'name': param['name'], 'type': 'string'}for param in nested_custom_task_special_params
                 ])
-                custom_task_cr['spec']['pipelineSpec']['params'] = sorted(
-                    custom_task_cr['spec']['pipelineSpec']['params'], key=lambda k: k['name'])
 
                 if nested_custom_task['ancestors']:
                     for custom_task_cr_again in custom_task_crs:
@@ -247,19 +277,35 @@ def _handle_tekton_custom_task(custom_task: dict, workflow: dict, recursive_task
                         if nested_param['name'] == special_param['name']:
                             nested_param['value'] = '$(params.%s)' % nested_param['name']
                 # need process parameters to replace results
+                custom_task_cr_task_names = [cr_task['name'] for cr_task in custom_task_cr['spec']['pipelineSpec']['tasks']]
                 for nested_custom_task_param in nested_custom_task_spec['params']:
                     if '$(tasks.' in nested_custom_task_param['value']:
-                        nested_custom_task_spec = json.loads(
-                            json.dumps(nested_custom_task_spec).replace(nested_custom_task_param['value'],
-                            '$(params.%s)' % nested_custom_task_param['name']))
+                        param_results = re.findall('\$\(tasks.([^ \t\n.:,;\{\}]+).results.([^ \t\n.:,;\{\}]+)\)',
+                                                    nested_custom_task_param['value'])
+                        for param_result in param_results:
+                            if param_result[0] not in custom_task_cr_task_names:
+                                custom_task_cr_param_names = [p['name'] for p in custom_task_cr['spec']['pipelineSpec']['params']]
+                                if nested_custom_task_param['name'] not in custom_task_cr_param_names:
+                                    for index, param in enumerate(nested_custom_task_spec['params']):
+                                        if nested_custom_task_param['name'] == param['name']:
+                                            nested_custom_task_spec['params'].pop(index)
+                                            break
+                                else:
+                                    nested_custom_task_spec = json.loads(
+                                        json.dumps(nested_custom_task_spec).replace(nested_custom_task_param['value'],
+                                        '$(params.%s)' % nested_custom_task_param['name']))
                 # add nested custom task spec to main custom task
                 custom_task_cr['spec']['pipelineSpec']['tasks'].append(nested_custom_task_spec)
+                custom_task_cr['spec']['pipelineSpec']['params'] = sorted(
+                    custom_task_cr['spec']['pipelineSpec']['params'], key=lambda k: k['name'])
 
     # remove the tasks belong to custom task from main workflow
     task_name_prefix = '-'.join(group_names[:-1] + [""])
     for task in tasks:
         if task['name'].replace(task_name_prefix, "") not in task_list:
-            new_tasks.append(task)
+            task_list_trimmed = [sanitize_k8s_name(task, max_length=LOOP_GROUP_NAME_LENGTH, rev_truncate=True) for task in task_list]
+            if task['name'].replace(task_name_prefix, "") not in task_list_trimmed:
+                new_tasks.append(task)
     workflow['spec']['pipelineSpec']['tasks'] = new_tasks
     return custom_task_crs, workflow
 
