@@ -12,21 +12,22 @@
 # See the License for the specific language governing permissions and
 # limitations under the License.
 
+import ast
 import inspect
 import json
-import tarfile
-import zipfile
-import re
-import textwrap
-import yaml
 import os
+import re
+import tarfile
+import textwrap
 import uuid
-import ast
-
-from typing import Callable, List, Text, Dict, Any
-from os import environ as env
+import zipfile
+from collections import defaultdict
 from distutils.util import strtobool
+import collections
+from os import environ as env
+from typing import Callable, List, Text, Dict, Any
 
+import yaml
 # Kubeflow Pipeline imports
 from kfp import dsl
 from kfp.compiler._default_transformers import add_pod_env
@@ -34,17 +35,15 @@ from kfp.compiler.compiler import Compiler
 from kfp.components.structures import InputSpec
 from kfp.dsl._for_loop import LoopArguments
 from kfp.dsl._metadata import _extract_pipeline_metadata
-from collections import defaultdict
-
 # KFP-Tekton imports
 from kfp_tekton.compiler import __tekton_api_version__ as tekton_api_version
 from kfp_tekton.compiler._data_passing_rewriter import fix_big_data_passing
 from kfp_tekton.compiler._k8s_helper import convert_k8s_obj_to_json, sanitize_k8s_name, sanitize_k8s_object
 from kfp_tekton.compiler._op_to_template import _op_to_template
-from kfp_tekton.compiler.yaml_utils import dump_yaml
-from kfp_tekton.compiler.pipeline_utils import TektonPipelineConf
 from kfp_tekton.compiler._tekton_handler import _handle_tekton_pipeline_variables, _handle_tekton_custom_task
-from kfp_tekton.tekton import TEKTON_CUSTOM_TASK_IMAGES, DEFAULT_CONDITION_OUTPUT_KEYWORD
+from kfp_tekton.compiler.pipeline_utils import TektonPipelineConf
+from kfp_tekton.compiler.yaml_utils import dump_yaml
+from kfp_tekton.tekton import TEKTON_CUSTOM_TASK_IMAGES, DEFAULT_CONDITION_OUTPUT_KEYWORD, LOOP_PIPELINE_NAME_LENGTH, LOOP_GROUP_NAME_LENGTH
 
 DEFAULT_ARTIFACT_BUCKET = env.get('DEFAULT_ARTIFACT_BUCKET', 'mlpipeline')
 DEFAULT_ARTIFACT_ENDPOINT = env.get('DEFAULT_ARTIFACT_ENDPOINT', 'minio-service.kubeflow:9000')
@@ -131,11 +130,15 @@ class TektonCompiler(Compiler):
     self._group_names = []
     self.pipeline_labels = {}
     self.pipeline_annotations = {}
+    self.tekton_inline_spec = True
+    self.resource_in_separate_yaml = False
     super().__init__(**kwargs)
 
   def _set_pipeline_conf(self, tekton_pipeline_conf: TektonPipelineConf):
     self.pipeline_labels = tekton_pipeline_conf.pipeline_labels
     self.pipeline_annotations = tekton_pipeline_conf.pipeline_annotations
+    self.tekton_inline_spec = tekton_pipeline_conf.tekton_inline_spec
+    self.resource_in_separate_yaml = tekton_pipeline_conf.resource_in_separate_yaml
 
   def _resolve_value_or_reference(self, value_or_reference, potential_references):
     """_resolve_value_or_reference resolves values and PipelineParams, which could be task parameters or input parameters.
@@ -180,10 +183,14 @@ class TektonCompiler(Compiler):
     """
     # Generate GroupOp template
     sub_group = group
-    # For loop and recursion usually append 16-19 characters, so limit the loop/recusion pipeline_name to 44 char
-    self._group_names = [sanitize_k8s_name(pipeline_name, max_length=44), sanitize_k8s_name(sub_group.name)]
+    # For loop and recursion id appends 5 characters, so limit the loop/recusion pipeline_name to 44 char and group_name to 12
+    # Group_name is truncated reversely because it has an unique identifier at the end of the name.
+    pipeline_name_copy = sanitize_k8s_name(pipeline_name, max_length=LOOP_PIPELINE_NAME_LENGTH)
+    sub_group_name_copy = sanitize_k8s_name(sub_group.name, max_length=LOOP_GROUP_NAME_LENGTH, rev_truncate=True)
+    self._group_names = [pipeline_name_copy, sub_group_name_copy]
     if self.uuid:
       self._group_names.insert(1, self.uuid)
+    # pipeline name (max 40) + loop id (max 5) + group name (max 16) + two connecting dashes (2) = 63 (Max size for CRD names)
     group_name = '-'.join(self._group_names) if group_type == "loop" or group_type == "graph" else sub_group.name
     template = {
       'metadata': {
@@ -218,7 +225,9 @@ class TektonCompiler(Compiler):
       # Special handling for recursive subgroup
       if sub_group.recursive_ref:
         # generate ref graph name
-        tmp_group_names = [pipeline_name, sanitize_k8s_name(sub_group.recursive_ref.name)]
+        sub_group_recursive_name_copy = sanitize_k8s_name(sub_group.recursive_ref.name,
+                                        max_length=LOOP_GROUP_NAME_LENGTH, rev_truncate=True)
+        tmp_group_names = [pipeline_name_copy, sub_group_recursive_name_copy]
         if self.uuid:
           tmp_group_names.insert(1, self.uuid)
         ref_group_name = '-'.join(tmp_group_names)
@@ -279,9 +288,7 @@ class TektonCompiler(Compiler):
                 self.loops_pipeline[group_name]['task_list'].append(sanitize_k8s_name(condition_op.name))
             if op.groups:
               for condition_op in op.groups:
-                # graph task should be a nested cr and need be appended to task list.
-                if condition_op.type == 'graph':
-                  self.loops_pipeline[group_name]['task_list'].append(sanitize_k8s_name(condition_op.name))
+                self.loops_pipeline[group_name]['task_list'].append(sanitize_k8s_name(condition_op.name))
         self.loops_pipeline[group_name]['spec']['name'] = group_name
         self.loops_pipeline[group_name]['spec']['taskRef'] = {
           "apiVersion": "custom.tekton.dev/v1alpha1",
@@ -373,7 +380,7 @@ class TektonCompiler(Compiler):
             sanitized_tasks.append(c_dict)
           loop_args_str_value = json.dumps(sanitized_tasks, sort_keys=True)
         else:
-          loop_args_str_value = str(loop_arg_value)
+          loop_args_str_value = json.dumps(loop_arg_value)
 
         self.loops_pipeline[group_name]['spec']['params'] = [{
           "name": sub_group.loop_args.full_name,
@@ -739,7 +746,7 @@ class TektonCompiler(Compiler):
             for index, item in enumerate(container_args):
               if item.startswith('--'):
                 custom_task_args[item[2:]] = container_args[index + 1]
-            non_param_keys = ['name', 'apiVersion', 'kind', 'taskSpec']
+            non_param_keys = ['name', 'apiVersion', 'kind', 'taskSpec', 'taskRef']
             task_params = []
             for key, value in custom_task_args.items():
               if key not in non_param_keys:
@@ -755,24 +762,42 @@ class TektonCompiler(Compiler):
                 'kind': custom_task_args['kind']
               }
             }
+            # Only one of --taskRef and --taskSpec allowed.
+            if custom_task_args.get('taskRef', '') and custom_task_args.get('taskSpec', ''):
+              raise("Custom task invalid configuration %s, Only one of --taskRef and --taskSpec allowed." % custom_task_args)
+            if custom_task_args.get('taskRef', ''):
+              try:
+                custom_task_cr = {
+                  'apiVersion': custom_task_args['apiVersion'],
+                  'kind': custom_task_args['kind'],
+                  'metadata': {
+                    'name': custom_task_args['name']
+                  },
+                  'spec': ast.literal_eval(custom_task_args['taskRef'])
+                }
+                for existing_cr in self.custom_task_crs:
+                  if existing_cr == custom_task_cr:
+                    # Skip duplicated CR resource
+                    custom_task_cr = {}
+                    break
+                if custom_task_cr:
+                  self.custom_task_crs.append(custom_task_cr)
+              except ValueError:
+                raise("Custom task ref %s is not a valid Python Dictionary" % custom_task_args['taskRef'])
+            # Setting --taskRef flag indicates, that spec be inlined.
             if custom_task_args.get('taskSpec', ''):
               try:
-                if custom_task_args['taskSpec']:
-                  custom_task_cr = {
+                task_ref = {
+                  'name': template['metadata']['name'],
+                  'params': task_params,
+                  # For processing Tekton parameter mapping later on.
+                  'orig_params': task_ref['params'],
+                  'taskSpec': {
                     'apiVersion': custom_task_args['apiVersion'],
                     'kind': custom_task_args['kind'],
-                    'metadata': {
-                      'name': custom_task_args['name']
-                    },
                     'spec': ast.literal_eval(custom_task_args['taskSpec'])
                   }
-                  for existing_cr in self.custom_task_crs:
-                    if existing_cr == custom_task_cr:
-                      # Skip duplicated CR resource
-                      custom_task_cr = {}
-                      break
-                  if custom_task_cr:
-                    self.custom_task_crs.append(custom_task_cr)
+                }
               except ValueError:
                 raise("Custom task spec %s is not a valid Python Dictionary" % custom_task_args['taskSpec'])
             # Pop custom task artifacts since we have no control of how
@@ -869,14 +894,25 @@ class TektonCompiler(Compiler):
     for loop_task_key in self.loops_pipeline.keys():
       task_name_prefix = '-'.join(self._group_names[:-1] + [""])
       raw_task_key = loop_task_key.replace(task_name_prefix, "")
+      for key in opgroup_name_to_parent_groups.keys():
+        if raw_task_key in key:
+          raw_task_key = key
+          break
       parent_group = opgroup_name_to_parent_groups.get(raw_task_key, [])
       if parent_group:
         if condition_refs.get(parent_group[-2], []):
           self.loops_pipeline[loop_task_key]['spec']['when'] = condition_refs.get(parent_group[-2], [])
-          for i, param in enumerate(self.loops_pipeline[loop_task_key]['spec']["params"]):
-            if param["value"] == condition_refs.get(parent_group[-2], [])[0]["input"]:
-              self.loops_pipeline[loop_task_key]['spec']["params"].pop(i)
-              break
+          # In nested recursive loop, the children of the loop pipeline can be both another loop
+          # and the self recursive loop. Thus, we cannot simply pop unnecessary params in one
+          # loop pipeline without verifying all the dependent parameters. Because nested recursion
+          # can have cycles, the DSL DAG may not represent the full view of all the dependent parameters.
+          # TODO: 1. Break any cycle in the nested recursion so it can represent as an acyclic graph.
+          #       2. Once the graph is acyclic, check all the children parameters in the loop_task
+          #          and pop the unnecessary parameters using the below logic.
+          # for i, param in enumerate(self.loops_pipeline[loop_task_key]['spec']["params"]):
+          #   if param["value"] == condition_refs.get(parent_group[-2], [])[0]["input"]:
+          #     self.loops_pipeline[loop_task_key]['spec']["params"].pop(i)
+          #     break
 
     # process input parameters from upstream tasks
     pipeline_param_names = [p['name'] for p in params]
@@ -938,10 +974,11 @@ class TektonCompiler(Compiler):
     # add timeout params to task_refs, instead of task.
     for task in task_refs:
       op = pipeline.ops.get(task['name'])
-      # Custom task doesn't support timeout feature
-      if task.get('taskSpec', ''):
+      if not (task.get('taskRef', '') != '' or task.get('taskSpec', '') != '' and 'apiVersion' in task['taskSpec']):
         if op != None and (not TEKTON_GLOBAL_DEFAULT_TIMEOUT or op.timeout):
           task['timeout'] = '%ds' % op.timeout
+      elif op.timeout > 0:  # i.e. incase of custom task, don't set timeout field by default.
+        task['timeout'] = '%ds' % op.timeout
 
     # handle resourceOp cases in pipeline
     self._process_resourceOp(task_refs, pipeline)
@@ -982,9 +1019,9 @@ class TektonCompiler(Compiler):
         'params': [{
             'name': p['name'],
             'value': p.get('default', '')
-          } for p in params],
+          } for p in sorted(params, key=lambda x: x['name'])],
         'pipelineSpec': {
-          'params': params,
+          'params': sorted(params, key=lambda x: x['name']),
           'tasks': task_refs + condition_task_refs,
           'finally': finally_tasks
         }
@@ -1198,7 +1235,7 @@ class TektonCompiler(Compiler):
     Args:
       pipeline_func: pipeline functions with @dsl.pipeline decorator.
       package_path: the output workflow tar.gz file path. for example, "~/a.tar.gz"
-      type_check: whether to enable the type check or not, default: False.
+      type_check: whether to enable the type check or not, default: True.
       pipeline_conf: PipelineConf instance. Can specify op transforms,
                      image pull secrets and other pipeline-level configuration options.
                      Overrides any configuration that may be set by the pipeline.
@@ -1309,18 +1346,85 @@ class TektonCompiler(Compiler):
     # Separate loop workflow from the main workflow
     if self.loops_pipeline:
       pipeline_loop_crs, workflow = _handle_tekton_custom_task(self.loops_pipeline, workflow, self.recursive_tasks, self._group_names)
+      inlined_as_taskSpec: List[Text] = []
+      recursive_tasks_names: List[Text] = [x['taskRef'].get('name', "") for x in self.recursive_tasks]
+      if self.tekton_inline_spec:
+        # Step 1. inline all the pipeline_loop_crs as they may refer to each other.
+        for i in range(len(pipeline_loop_crs)):
+          if 'pipelineSpec' in pipeline_loop_crs[i]['spec']:
+            if 'params' in pipeline_loop_crs[i]['spec']['pipelineSpec']:
+              # Preserve order of params, required by tests.
+              pipeline_loop_crs[i]['spec']['pipelineSpec']['params'] =\
+                sorted(pipeline_loop_crs[i]['spec']['pipelineSpec']['params'], key=lambda kv: (kv['name']))
+            t, e = TektonCompiler._inline_tasks(pipeline_loop_crs[i]['spec']['pipelineSpec']['tasks'],
+                                                pipeline_loop_crs, recursive_tasks_names)
+            if e:
+              pipeline_loop_crs[i]['spec']['pipelineSpec']['tasks'] = t
+              inlined_as_taskSpec.extend(e)
+        # Step 2. inline pipeline_loop_crs in the workflow
+        workflow_tasks, e = TektonCompiler._inline_tasks(workflow['spec']['pipelineSpec']['tasks'],
+                                                         pipeline_loop_crs, recursive_tasks_names)
+        inlined_as_taskSpec.extend(e)
+        workflow['spec']['pipelineSpec']['tasks'] = workflow_tasks
+        # Preserve order of params, required by tests.
+        if 'params' in workflow['spec']:
+          workflow['spec']['params'] = sorted(workflow['spec']['params'], key=lambda kv: (kv['name']))
       TektonCompiler._write_workflow(workflow=workflow, package_path=package_path)
+
+      # create cr yaml for only those pipelineLoop cr which could not be converted to inlined spec.
+      loop_package_annotations = []
       for i in range(len(pipeline_loop_crs)):
-        TektonCompiler._write_workflow(workflow=pipeline_loop_crs[i],
-                                       package_path=os.path.splitext(package_path)[0] + "_pipelineloop_cr" + str(i + 1) + '.yaml')
+        if pipeline_loop_crs[i]['metadata'].get('name', "") not in inlined_as_taskSpec:
+          if self.resource_in_separate_yaml:
+            TektonCompiler._write_workflow(workflow=pipeline_loop_crs[i],
+                                           package_path=os.path.splitext(package_path)[0] +
+                                                        "_pipelineloop_cr" + str(i + 1) + '.yaml')
+          else:
+            pipeline_loop_cr = TektonCompiler._write_workflow(workflow=collections.OrderedDict(pipeline_loop_crs[i]))
+            loop_package_annotations.append(yaml.load(pipeline_loop_cr, Loader=yaml.FullLoader))
+      if loop_package_annotations:
+        workflow['metadata']['annotations']['tekton.dev/resource_templates'] = json.dumps(loop_package_annotations, sort_keys=True)
+      # Need to compiles after all the CRs being processed.
+      TektonCompiler._write_workflow(workflow=workflow, package_path=package_path)
     else:
       TektonCompiler._write_workflow(workflow=workflow, package_path=package_path)   # Tekton change
     # Separate custom task CR from the main workflow
     for i in range(len(self.custom_task_crs)):
       TektonCompiler._write_workflow(workflow=self.custom_task_crs[i],
-                                     package_path=os.path.splitext(package_path)[0] + "_customtask_cr" + str(i + 1) + '.yaml')
+                                    package_path=os.path.splitext(package_path)[0] +
+                                                  "_customtask_cr" + str(i + 1) + '.yaml')
     _validate_workflow(workflow)
 
+  @staticmethod
+  def _inline_tasks(tasks: List[Dict[Text, Any]], crs: List[Dict[Text, Any]], recursive_tasks: List[Text]):
+    """
+      Scan all the `tasks` and for each taskRef in `tasks` resolve it in `crs`
+       and inline them as taskSpec.
+       return tasks with all the taskRef -> taskSpec resolved.
+       list of names of the taskRef that were successfully converted.
+    """
+    workflow_tasks = tasks.copy()
+    inlined_as_taskSpec = []
+    for j in range(len(workflow_tasks)):
+      if 'params' in workflow_tasks[j]:
+        # Preserve order of params, required by tests.
+        workflow_tasks[j]['params'] = sorted(workflow_tasks[j]['params'], key=lambda kv: (kv['name']))
+      if 'taskRef' in workflow_tasks[j]:
+        wf_taskRef = workflow_tasks[j]['taskRef']
+        if 'name' in wf_taskRef and \
+                wf_taskRef['name'] not in recursive_tasks:  # we do not inline recursive tasks.
+          cr_apiVersion = wf_taskRef['apiVersion']
+          cr_kind = wf_taskRef['kind']
+          cr_ref_name = wf_taskRef['name']
+          for i in range(len(crs)):
+            if crs[i]['metadata'].get('name', "") == cr_ref_name:
+              workflow_tasks[j]['taskSpec'] = \
+                {'apiVersion': cr_apiVersion, 'kind': cr_kind,
+                 'spec': crs[i]['spec']}
+              inlined_as_taskSpec.append(cr_ref_name)
+              workflow_tasks[j].pop('taskRef')
+    return workflow_tasks, inlined_as_taskSpec
+
 
 def _validate_workflow(workflow: Dict[Text, Any]):
 
