@@ -12,12 +12,14 @@
 # See the License for the specific language governing permissions and
 # limitations under the License.
 
-import kfp.compiler.main as kfp_compiler_main
 import argparse
-import sys
 import os
-from . import TektonCompiler
+import sys
 
+import kfp.compiler.main as kfp_compiler_main
+from kfp_tekton.compiler.pipeline_utils import TektonPipelineConf
+
+from . import TektonCompiler
 from .. import __version__
 
 
@@ -44,12 +46,15 @@ def parse_arguments():
   parser.add_argument('--disable-type-check',
                       action='store_true',
                       help='disable the type check, default is enabled.')
+  parser.add_argument('--disable-task-inline',
+                      action='store_true',
+                      help='disable taskSpec inlining, default is enabled.')
 
   args = parser.parse_args()
   return args
 
 
-def _compile_pipeline_function(pipeline_funcs, function_name, output_path, type_check):
+def _compile_pipeline_function(pipeline_funcs, function_name, output_path, type_check, tekton_pipeline_conf=None):
   if len(pipeline_funcs) == 0:
     raise ValueError('A function with @dsl.pipeline decorator is required in the py file.')
 
@@ -65,16 +70,16 @@ def _compile_pipeline_function(pipeline_funcs, function_name, output_path, type_
   else:
     pipeline_func = pipeline_funcs[0]
 
-  TektonCompiler().compile(pipeline_func, output_path, type_check)
+  TektonCompiler().compile(pipeline_func, output_path, type_check, tekton_pipeline_conf=tekton_pipeline_conf)
 
 
-def compile_pyfile(pyfile, function_name, output_path, type_check):
+def compile_pyfile(pyfile, function_name, output_path, type_check, tekton_pipeline_conf=None):
   sys.path.insert(0, os.path.dirname(pyfile))
   try:
     filename = os.path.basename(pyfile)
     with kfp_compiler_main.PipelineCollectorContext() as pipeline_funcs:
       __import__(os.path.splitext(filename)[0])
-    _compile_pipeline_function(pipeline_funcs, function_name, output_path, type_check)
+    _compile_pipeline_function(pipeline_funcs, function_name, output_path, type_check, tekton_pipeline_conf)
   finally:
     del sys.path[0]
 
@@ -86,7 +91,11 @@ def main():
         (args.py is not None and args.package is not None)):
         raise ValueError('Either --py or --package is needed but not both.')
     if args.py:
-        compile_pyfile(args.py, args.function, args.output, not args.disable_type_check)
+        tekton_pipeline_conf = TektonPipelineConf()
+        tekton_pipeline_conf.set_tekton_inline_spec(True)
+        if args.disable_task_inline:
+            tekton_pipeline_conf.set_tekton_inline_spec(False)
+        compile_pyfile(args.py, args.function, args.output, not args.disable_type_check, tekton_pipeline_conf)
     else:
         if args.namespace is None:
             raise ValueError('--namespace is required for compiling packages.')
